from typing import Dict, List, Optional, Union

from pylabrobot.resources.plate import Plate

from .backend import (
  Calibrate,
  CarriageSpeed,
  KineticSettings,
  MolecularDevicesBackend,
  MolecularDevicesSettings,
  PmtGain,
  ReadOrder,
  ReadType,
  ShakeSettings,
  SpectrumSettings,
)


class MolecularDevicesSpectraMax384PlusBackend(MolecularDevicesBackend):
  """Backend for Molecular Devices SpectraMax 384 Plus plate readers."""

  def __init__(self, port: str) -> None:
    super().__init__(port)

  async def _set_readtype(self, settings: MolecularDevicesSettings) -> None:
    """Set the READTYPE command and the expected number of response fields."""
    cmd = f"!READTYPE {'CUV' if settings.cuvette else 'PLA'}"
    await self.send_command(cmd, num_res_fields=1)

  async def _set_nvram(self, settings: MolecularDevicesSettings) -> None:
    pass

  async def _set_tag(self, settings: MolecularDevicesSettings) -> None:
    pass

  async def read_fluorescence(  # type: ignore[override]
    self,
    plate: "Plate",
    excitation_wavelengths: List[int],
    emission_wavelengths: List[int],
    cutoff_filters: List[int],
    read_type: ReadType = ReadType.ENDPOINT,
    read_order: ReadOrder = ReadOrder.COLUMN,
    calibrate: Calibrate = Calibrate.ONCE,
    shake_settings: Optional["ShakeSettings"] = None,
    carriage_speed: CarriageSpeed = CarriageSpeed.NORMAL,
    read_from_bottom: bool = False,
    pmt_gain: Union[PmtGain, int] = PmtGain.AUTO,
    flashes_per_well: int = 10,
    kinetic_settings: Optional["KineticSettings"] = None,
    spectrum_settings: Optional["SpectrumSettings"] = None,
    cuvette: bool = False,
    settling_time: int = 0,
    timeout: int = 600,
  ) -> List[Dict]:
    raise NotImplementedError("Fluorescence reading is not supported.")

  async def read_luminescence(  # type: ignore[override]
    self,
    plate: "Plate",
    emission_wavelengths: List[int],
    read_type: ReadType = ReadType.ENDPOINT,
    read_order: ReadOrder = ReadOrder.COLUMN,
    calibrate: Calibrate = Calibrate.ONCE,
    shake_settings: Optional["ShakeSettings"] = None,
    carriage_speed: CarriageSpeed = CarriageSpeed.NORMAL,
    read_from_bottom: bool = False,
    pmt_gain: Union[PmtGain, int] = PmtGain.AUTO,
    flashes_per_well: int = 0,
    kinetic_settings: Optional["KineticSettings"] = None,
    spectrum_settings: Optional["SpectrumSettings"] = None,
    cuvette: bool = False,
    settling_time: int = 0,
    timeout: int = 600,
  ) -> List[Dict]:
    raise NotImplementedError("Luminescence reading is not supported.")

  async def read_fluorescence_polarization(
    self,
    plate: "Plate",
    excitation_wavelengths: List[int],
    emission_wavelengths: List[int],
    cutoff_filters: List[int],
    read_type: ReadType = ReadType.ENDPOINT,
    read_order: ReadOrder = ReadOrder.COLUMN,
    calibrate: Calibrate = Calibrate.ONCE,
    shake_settings: Optional["ShakeSettings"] = None,
    carriage_speed: CarriageSpeed = CarriageSpeed.NORMAL,
    read_from_bottom: bool = False,
    pmt_gain: Union[PmtGain, int] = PmtGain.AUTO,
    flashes_per_well: int = 10,
    kinetic_settings: Optional["KineticSettings"] = None,
    spectrum_settings: Optional["SpectrumSettings"] = None,
    cuvette: bool = False,
    settling_time: int = 0,
    timeout: int = 600,
  ) -> List[Dict]:
    raise NotImplementedError("Fluorescence polarization reading is not supported.")

  async def read_time_resolved_fluorescence(
    self,
    plate: "Plate",
    excitation_wavelengths: List[int],
    emission_wavelengths: List[int],
    cutoff_filters: List[int],
    delay_time: int,
    integration_time: int,
    read_type: ReadType = ReadType.ENDPOINT,
    read_order: ReadOrder = ReadOrder.COLUMN,
    calibrate: Calibrate = Calibrate.ONCE,
    shake_settings: Optional["ShakeSettings"] = None,
    carriage_speed: CarriageSpeed = CarriageSpeed.NORMAL,
    read_from_bottom: bool = False,
    pmt_gain: Union[PmtGain, int] = PmtGain.AUTO,
    flashes_per_well: int = 50,
    kinetic_settings: Optional["KineticSettings"] = None,
    spectrum_settings: Optional["SpectrumSettings"] = None,
    cuvette: bool = False,
    settling_time: int = 0,
    timeout: int = 600,
  ) -> List[Dict]:
    raise NotImplementedError("Time-resolved fluorescence reading is not supported.")



import warnings


class ResourceNotFoundError(Exception):
  pass


class TooLittleLiquidError(Exception):
  """Raised when trying to aspirate more liquid from a container than is still present."""


class TooLittleVolumeError(Exception):
  """Raised when trying to dispense more liquid into a container than is still available."""


class HasTipError(Exception):
  """Raised when a tip already exists in a location where a tip is being added."""


class NoTipError(Exception):
  """Raised when a tip was expected but none was found."""


class CrossContaminationError(Exception):
  """Raised when attempting to aspirate from a well with a tip that has touched a
  different liquid."""

  def __init__(self, *args, **kwargs):
    warnings.warn(
      "Cross contamination tracking is deprecated and will be removed in a future version. ",
      DeprecationWarning,
      stacklevel=2,
    )
    super().__init__(*args, **kwargs)


class ResourceDefinitionIncompleteError(Exception):
  """Raised when trying to access a resource that has not been defined or is not complete.

  We have some "phantom" resources that have a name and creator function, but are missing some
  information, or they don't have enough metadata to uniquely identify them. This means they are
  effectively useless. These resources often originate from a database import (like venus) that is
  incomplete.

  This error is raised when you try to create a resource like that. Please create a PR to list the
  resource catalog number (or equivalent), or measure and contribute the missing information. Please
  create an issue if you need help with this.

  Tracking the general problem in https://github.com/PyLabRobot/pylabrobot/issues/170.
  """

  def __init__(self, resource_name: str):
    super().__init__(
      f"Resource '{resource_name}' is incomplete and cannot be used. "
      "Please create a PR to complete this resource, or create an issue if you "
      "need help. https://github.com/PyLabRobot/pylabrobot"
    )


class NoLocationError(Exception):
  """Raised when trying to access a location that does not exist, eg. when calling
  get_absolute_location on a resource where some resource between it and the root
  has no location."""



import asyncio
import logging
import ssl
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from pylabrobot.io.capture import Command, capturer, get_capture_or_validation_active
from pylabrobot.io.errors import ValidationError
from pylabrobot.io.io import IOBase
from pylabrobot.io.validation_utils import LOG_LEVEL_IO, align_sequences

if TYPE_CHECKING:
  from pylabrobot.io.capture import CaptureReader


logger = logging.getLogger(__name__)


@dataclass
class SocketCommand(Command):
  data: str

  def __init__(self, device_id: str, action: str, data: str, module: str = "socket"):
    super().__init__(module=module, device_id=device_id, action=action)
    self.data = data


class Socket(IOBase):
  """IO for reading/writing to a TCP socket."""

  def __init__(
    self,
    human_readable_device_name: str,
    host: str,
    port: int,
    read_timeout: float = 30,
    write_timeout: float = 30,
    ssl_context: Optional[ssl.SSLContext] = None,
    server_hostname: Optional[str] = None,
  ):
    self._human_readable_device_name = human_readable_device_name
    self._host = host
    self._port = port
    self._reader: Optional[asyncio.StreamReader] = None
    self._writer: Optional[asyncio.StreamWriter] = None
    self._read_timeout = read_timeout
    self._write_timeout = write_timeout
    self._ssl_context = ssl_context
    self._server_hostname = server_hostname
    self._unique_id = f"{self._host}:{self._port}"
    self._read_lock = asyncio.Lock()
    self._write_lock = asyncio.Lock()
    self._ssl = ssl

    if get_capture_or_validation_active():
      raise RuntimeError("Cannot create a new Socket object while capture or validation is active")

  async def setup(self):
    await self._connect()

  async def _connect(self):
    self._reader, self._writer = await asyncio.open_connection(
      host=self._host,
      port=self._port,
      ssl=self._ssl_context,
      server_hostname=self._server_hostname,
    )

  async def stop(self):
    await self._disconnect()

  async def _disconnect(self):
    async with self._read_lock, self._write_lock:
      self._reader = None
      if self._writer is None:
        return

      logger.info("Closing connection to socket %s:%s", self._host, self._port)

      try:
        self._writer.close()
        await self._writer.wait_closed()
      except OSError as e:
        logger.warning("Error while closing socket connection: %s", e)
      finally:
        self._writer = None

  async def reconnect(self):
    await self._disconnect()
    await self._connect()

  def serialize(self):
    return {
      "human_readable_device_name": self._human_readable_device_name,
      "host": self._host,
      "port": self._port,
      "type": "Socket",
      "read_timeout": self._read_timeout,
      "write_timeout": self._write_timeout,
    }

  @classmethod
  def deserialize(cls, data: dict) -> "Socket":
    kwargs = {}
    if "read_timeout" in data:
      kwargs["read_timeout"] = data["read_timeout"]
    if "write_timeout" in data:
      kwargs["write_timeout"] = data["write_timeout"]
    return cls(
      human_readable_device_name=data["human_readable_device_name"],
      host=data["host"],
      port=data["port"],
      **kwargs,
    )

  async def write(self, data: bytes, timeout: Optional[float] = None) -> None:
    """Wrapper around StreamWriter.write with lock and io logging.
    Does not retry on timeouts.
    """
    if self._writer is None:
      raise RuntimeError(
        f"Socket for '{self._human_readable_device_name}' not set up; call setup() first"
      )
    timeout = self._write_timeout if timeout is None else timeout
    async with self._write_lock:
      self._writer.write(data)
      logger.log(LOG_LEVEL_IO, "[%s:%d] write %s", self._host, self._port, data)
      capturer.record(
        SocketCommand(
          device_id=self._unique_id,
          action="write",
          data=data.hex(),
        )
      )
      try:
        await asyncio.wait_for(self._writer.drain(), timeout=timeout)
        return
      except asyncio.TimeoutError as exc:
        logger.error("write timeout: %r", exc)
        raise TimeoutError(f"Timeout while writing to socket after {timeout} seconds") from exc
      except (ConnectionResetError, OSError) as e:
        logger.error("write error: %r", e)
        raise

  async def read(self, num_bytes: int = 128, timeout: Optional[float] = None) -> bytes:
    """Wrapper around StreamReader.read with lock and io logging.

    Args:
      num_bytes: The maximum number of bytes to read from the socket.
        If fewer bytes are available, the method may return less than `num_bytes`.
        If the end of the stream is reached before `num_bytes` bytes are read, only the available bytes are returned.
      timeout: Maximum time to wait for data before raising a timeout.

    Returns:
      The data read from the socket, which may be fewer than `num_bytes` bytes.
    """
    if self._reader is None:
      raise RuntimeError(
        f"Socket for '{self._human_readable_device_name}' not set up; call setup() first"
      )
    timeout = self._read_timeout if timeout is None else timeout
    async with self._read_lock:
      try:
        data = await asyncio.wait_for(self._reader.read(num_bytes), timeout=timeout)
      except asyncio.TimeoutError as exc:
        logger.error("read timeout: %r", exc)
        raise TimeoutError(f"Timeout while reading from socket after {timeout} seconds") from exc
      logger.log(LOG_LEVEL_IO, "[%s:%d] read %s", self._host, self._port, data)
      capturer.record(
        SocketCommand(
          device_id=self._unique_id,
          action="read",
          data=data.hex(),
        )
      )
      return data

  async def readline(self, timeout: Optional[float] = None) -> bytes:
    """Wrapper around StreamReader.readline with lock and io logging."""
    if self._reader is None:
      raise RuntimeError(
        f"Socket for '{self._human_readable_device_name}' not set up; call setup() first"
      )
    timeout = self._read_timeout if timeout is None else timeout
    async with self._read_lock:
      try:
        data = await asyncio.wait_for(self._reader.readline(), timeout=timeout)
      except asyncio.TimeoutError as exc:
        logger.error("readline timeout: %r", exc)
        raise TimeoutError(f"Timeout while reading from socket after {timeout} seconds") from exc
      logger.log(LOG_LEVEL_IO, "[%s:%d] readline %s", self._host, self._port, data)
      capturer.record(
        SocketCommand(
          device_id=self._unique_id,
          action="readline",
          data=data.hex(),
        )
      )
      return data

  async def readuntil(self, separator: bytes = b"\n", timeout: Optional[float] = None) -> bytes:
    """Wrapper around StreamReader.readuntil with lock and io logging.
    Do not retry on timeouts."""
    if self._reader is None:
      raise RuntimeError(
        f"Socket for '{self._human_readable_device_name}' not set up; call setup() first"
      )
    timeout = self._read_timeout if timeout is None else timeout
    async with self._read_lock:
      try:
        data = await asyncio.wait_for(self._reader.readuntil(separator), timeout=timeout)
      except asyncio.TimeoutError as exc:
        logger.error("readuntil timeout: %r", exc)
        raise TimeoutError(f"Timeout while reading from socket after {timeout} seconds") from exc
      logger.log(LOG_LEVEL_IO, "[%s:%d] readuntil %s", self._host, self._port, data)
      capturer.record(
        SocketCommand(
          device_id=self._unique_id,
          action="readuntil:" + separator.hex(),
          data=data.hex(),
        )
      )
      return data

  async def read_exact(self, num_bytes: int, timeout: Optional[float] = None) -> bytes:
    """Read exactly num_bytes, blocking until all bytes are received.

    Args:
      num_bytes: The exact number of bytes to read.
      timeout: Maximum time to wait for data before raising a timeout.
        Note: The timeout is applied per-chunk read operation, not cumulatively
        for the entire read. For small reads (typical use case), this is acceptable.
        For large reads, consider that the total time may exceed the timeout value.

    Returns:
      Exactly num_bytes of data.

    Raises:
      ConnectionError: If the connection is closed before num_bytes are read.
      TimeoutError: If timeout is reached before num_bytes are read.
    """
    if self._reader is None:
      raise RuntimeError(
        f"Socket for '{self._human_readable_device_name}' not set up; call setup() first"
      )
    timeout = self._read_timeout if timeout is None else timeout
    data = bytearray()
    async with self._read_lock:
      while len(data) < num_bytes:
        remaining = num_bytes - len(data)
        try:
          chunk = await asyncio.wait_for(self._reader.read(remaining), timeout=timeout)
        except asyncio.TimeoutError as exc:
          logger.error("read_exact timeout: %r", exc)
          raise TimeoutError(f"Timeout while reading from socket after {timeout} seconds") from exc
        if len(chunk) == 0:
          raise ConnectionError("Connection closed before num_bytes are read")
        data.extend(chunk)

      result = bytes(data)
      logger.log(LOG_LEVEL_IO, "[%s:%d] read_exact %s", self._host, self._port, result.hex())
      capturer.record(
        SocketCommand(
          device_id=self._unique_id,
          action="read_exact",
          data=result.hex(),
        )
      )
      return result

  async def read_until_eof(self, chunk_size: int = 1024, timeout: Optional[float] = None) -> bytes:
    """Read until EOF is reached.
    Do not retry on timeouts.
    """
    buf = bytearray()
    timeout = self._read_timeout if timeout is None else timeout

    async with self._read_lock:
      while True:
        if self._reader is None:
          raise RuntimeError(
            f"Socket for '{self._human_readable_device_name}' not set up; call setup() first"
          )
        try:
          chunk = await asyncio.wait_for(self._reader.read(chunk_size), timeout=timeout)
        except asyncio.TimeoutError as exc:
          # if some previous read attempts already return some data, we should consider this a success
          if len(buf) > 0:
            break
          logger.error("read_until_eof timeout: %r", exc)
          raise TimeoutError(f"Timeout while reading from socket after {timeout} seconds") from exc
        if len(chunk) == 0:
          break

        logger.debug("read_until_eof: got %d bytes", len(chunk))
        buf.extend(chunk)

      line = bytes(buf)
      logger.log(LOG_LEVEL_IO, "[%s:%d] read_until_eof %s", self._host, self._port, line)
      capturer.record(
        SocketCommand(
          device_id=self._unique_id,
          action="read_until_eof",
          data=line.hex(),
        )
      )
      return line


class SocketValidator(Socket):
  """Socket validator for testing/validation purposes."""

  def __init__(
    self,
    cr: "CaptureReader",
    human_readable_device_name: str,
    host: str,
    port: int,
    read_timeout: float = 30,
    write_timeout: float = 30,
  ):
    super().__init__(
      human_readable_device_name=human_readable_device_name,
      host=host,
      port=port,
      read_timeout=read_timeout,
      write_timeout=write_timeout,
    )
    self.cr = cr

  async def setup(self):
    """Mock setup for validation."""

  async def stop(self):
    """Mock stop for validation."""
    return

  async def _connect(self):
    """Mock connect for validation."""

  async def _disconnect(self):
    """Mock disconnect for validation."""

  async def write(self, data: bytes, *args, **kwargs):
    """Validate write command against captured data."""
    next_command = SocketCommand(**self.cr.next_command())
    if not (
      next_command.module == "socket"
      and next_command.device_id == self._unique_id
      and next_command.action == "write"
    ):
      raise ValidationError(
        f"Expected socket write command to {self._unique_id}, "
        f"got {next_command.module} {next_command.action} to {next_command.device_id}"
      )
    expected = bytes.fromhex(next_command.data)
    if not expected == data:
      align_sequences(expected=expected.decode("latin-1"), actual=data.decode("latin-1"))
      raise ValidationError("Data mismatch: difference was written to stdout.")

  async def read(self, *args, **kwargs) -> bytes:
    """Return captured read data for validation."""
    next_command = SocketCommand(**self.cr.next_command())
    if not (
      next_command.module == "socket"
      and next_command.device_id == self._unique_id
      and next_command.action == "read"
    ):
      raise ValidationError(
        f"Expected socket read command from {self._unique_id}, "
        f"got {next_command.module} {next_command.action} from {next_command.device_id}"
      )
    return bytes.fromhex(next_command.data)

  async def readline(self, *args, **kwargs) -> bytes:
    """Return captured readline data for validation."""
    next_command = SocketCommand(**self.cr.next_command())
    if not (
      next_command.module == "socket"
      and next_command.device_id == self._unique_id
      and next_command.action == "readline"
    ):
      raise ValidationError(
        f"Expected socket readline command from {self._unique_id}, "
        f"got {next_command.module} {next_command.action} from {next_command.device_id}"
      )
    return bytes.fromhex(next_command.data)

  async def readuntil(self, separator: bytes = b"\n", *args, **kwargs) -> bytes:
    """Return captured readuntil data for validation."""
    next_command = SocketCommand(**self.cr.next_command())
    if not (
      next_command.module == "socket"
      and next_command.device_id == self._unique_id
      and next_command.action == "readuntil:" + separator.hex()
    ):
      expected_sep: Optional[bytes]
      if "readuntil:" in next_command.action:
        expected_sep = bytes.fromhex(next_command.action.split("readuntil:")[1])
      else:
        expected_sep = None
      raise ValidationError(
        f"Expected socket readuntil command from {self._unique_id}, "
        f"got {next_command.module} {next_command.action} from {next_command.device_id}"
        f" (expected separator {expected_sep!r}, got {separator!r})"
      )
    return bytes.fromhex(next_command.data)

  async def read_exact(self, *args, **kwargs) -> bytes:
    """Return captured read_exact data for validation."""
    next_command = SocketCommand(**self.cr.next_command())
    if not (
      next_command.module == "socket"
      and next_command.device_id == self._unique_id
      and next_command.action == "read_exact"
    ):
      raise ValidationError(
        f"Expected socket read_exact command from {self._unique_id}, "
        f"got {next_command.module} {next_command.action} from {next_command.device_id}"
      )
    return bytes.fromhex(next_command.data)

  async def read_until_eof(self, *args, **kwargs) -> bytes:
    """Return captured read_until_eof data for validation."""
    next_command = SocketCommand(**self.cr.next_command())
    if not (
      next_command.module == "socket"
      and next_command.device_id == self._unique_id
      and next_command.action == "read_until_eof"
    ):
      raise ValidationError(
        f"Expected socket read_until_eof command from {self._unique_id}, "
        f"got {next_command.module} {next_command.action} from {next_command.device_id}"
      )
    return bytes.fromhex(next_command.data)



import asyncio

from pylabrobot.machines.machine import Machine

from .backend import FanBackend


class Fan(Machine):
  """
  Front end for Fans.
  """

  def __init__(self, backend: FanBackend):
    super().__init__(backend=backend)
    self.backend: FanBackend = backend  # fix type

  async def stop(self):
    await self.backend.turn_off()
    await super().stop()

  async def turn_on(self, intensity: int, duration=None):
    """Run the fan

    Args:
      intensity: integer percent between 0 and 100
      duration: time to run the fan for. If None, run until `turn_off` is called.
    """

    await self.backend.turn_on(intensity=intensity)

    if duration is not None:
      await asyncio.sleep(duration)
      await self.backend.turn_off()

  async def turn_off(self):
    """Turn the fan off, but do not close the connection."""
    await self.backend.turn_off()



"""Config module. Facilitates reading and writing module-level config files.
Checks the current directory and all parent directories for a `pylabrobot.ini`
config file. If the config file does not exist, a default Config object will be
created and written to the project directory containing the .git directory. If
this does not exist, the current directory will be used.
"""

from pathlib import Path
from typing import Optional, Union

from pylabrobot.config.config import Config
from pylabrobot.config.formats import MultiLoader
from pylabrobot.config.formats.ini_config import IniLoader, IniSaver
from pylabrobot.config.formats.json_config import JsonLoader
from pylabrobot.config.io.file import FileReader, FileWriter

DEFAULT_LOADERS = [IniLoader(), JsonLoader()]

DEFAULT_CONFIG_READER = FileReader(
  format_loader=MultiLoader(DEFAULT_LOADERS),
)

DEFAULT_CONFIG_WRITER = FileWriter(
  format_saver=IniSaver(),
)


def get_file(base_name: str, _dir: Path) -> Optional[Path]:
  for ext in (rdr.extension for rdr in DEFAULT_LOADERS):
    cfg = _dir / f"{base_name}.{ext}"
    if cfg.exists():
      return cfg
  return None


def get_config_file(base_name: str, cur_dir: Optional[Union[str, Path]] = None) -> Optional[Path]:
  """Get the path to the config file.

  Args:
    base_name: The base name of the config file.

  Returns:
    The path to the config file.
  """
  cdir = Path(cur_dir) if cur_dir is not None else Path.cwd()

  cfg = get_file(base_name, cdir)
  if cfg is not None:
    return cfg

  if cdir.parent == cdir:
    return None

  return get_config_file(base_name, cdir.parent)


def get_dir_to_create_config_file_in() -> Path:
  """Crawls parent directories and looks for a .git directory to determine the
  root of the project. If no .git directory is found, the current directory is
  returned."""
  cur_dir = Path.cwd()
  for parent in cur_dir.parents:
    if (parent / ".git").exists():
      return parent
  return cur_dir


def load_config(
  base_file_name: str,
  create_default: bool = False,
  create_module_level: bool = True,
) -> Config:
  """Load a Config object from a file.

  Args:
    base_file_name: The base file name to load.
    create_default: Whether to create a default Config object if the file does
    not exist. It will be created with file extension that is first in
    DEFAULT_READER.
    create_module_level: Whether to create the default file in the module level
  """
  config_path = get_config_file(base_file_name)
  if config_path is None:
    if not create_default:
      return Config()
    create_dir = get_dir_to_create_config_file_in() if create_module_level else Path.cwd()
    default_extension = list(rdr.extension for rdr in DEFAULT_LOADERS)[0]
    config_path = create_dir / f"{base_file_name}.{default_extension}"
    DEFAULT_CONFIG_WRITER.write(config_path, Config())

  return DEFAULT_CONFIG_READER.read(config_path)



from abc import ABCMeta, abstractmethod

from pylabrobot.machines.backend import MachineBackend


class ScaleBackend(MachineBackend, metaclass=ABCMeta):
  """Backend for a scale"""

  @abstractmethod
  async def zero(self): ...

  @abstractmethod
  async def tare(self): ...

  @abstractmethod
  async def read_weight(self) -> float:
    """Read the weight in grams"""
    ...

  # Deprecated: for backward compatibility
  async def get_weight(self) -> float:
    """Deprecated: Use read_weight() instead.

    Get the weight in grams"""
    import warnings

    warnings.warn(
      "get_weight() is deprecated and will be removed in 2026-03. Use read_weight() instead.",
      DeprecationWarning,
      stacklevel=2,
    )
    return await self.read_weight()



import logging
from typing import Optional

from .base_control import BaseControl
from .spark_enums import InstrumentMessageType, ModuleType


class DataControl(BaseControl):
  async def get_programmable_memory_scope(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the programmable memory scope."""
    command = "#DOWNLOAD TYPE"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    response = await self.send_command(command)
    return response

  async def get_ranges_for_memory_scope(
    self,
    memory_scope: str,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the ranges for a memory scope."""
    command = "#DOWNLOAD"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    command += f" TYPE={memory_scope}"
    return await self.send_command(command)

  async def prepare_download(
    self,
    memory_scope: str,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Prepares the device for download."""
    command = "DOWNLOAD PREPARE"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    command += f" TYPE={memory_scope}"
    return await self.send_command(command)

  async def start_download_block(
    self,
    offset: int,
    size: int,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Starts downloading a block of data."""
    command = "DOWNLOAD BLOCK START"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    command += f" OFFSET={offset} SIZE={size}"
    return await self.send_command(command)

  async def end_download_block(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Ends the download of a block of data."""
    command = "DOWNLOAD BLOCK END"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_download_sections(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the available download sections."""
    command = "#DOWNLOAD SECTION NAME"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    response = await self.send_command(command)
    return response

  async def start_download_section(
    self,
    section_name: str,
    size: int,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Starts downloading a specific section."""
    command = "DOWNLOAD SECTION START"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    command += f" NAME={section_name} SIZE={size}"
    return await self.send_command(command)

  async def end_download_section(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Ends the download of a section."""
    command = "DOWNLOAD SECTION END"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def finalize_download(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Finalizes the download process."""
    command = "DOWNLOAD FINISH"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def send_download_data(self, data: bytes) -> None:
    """Sends data to the device during download."""
    raise NotImplementedError
    # return await self._usb_write(self.ep_bulk_out, data)

  async def get_command_buffer_size(self) -> Optional[str]:
    """Gets the command buffer size."""
    return await self.send_command("?BUFFER COMMAND SIZE")

  async def get_command_overhead(self) -> Optional[str]:
    """Gets the command overhead."""
    return await self.send_command("?BUFFER COMMAND OVERHEAD")

  async def clear_error_stack(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Clears the error stack."""
    command = "LASTERROR CLEAR"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_error_index_range(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the error index range."""
    command = "#LASTERROR INDEX"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_current_max_error_index(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the current maximum error index."""
    command = "?LASTERROR MAX"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_error(
    self,
    index: int,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the error at the given index."""
    command = f"?LASTERROR INDEX={index}"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_module_sap_number(self) -> Optional[str]:
    """Gets the module SAP number."""
    return await self.send_command("?INFO SAP_NR_MODULE")

  async def get_instrument_sap_number(self) -> Optional[str]:
    """Gets the instrument SAP number."""
    return await self.send_command("?INFO SAP_NR_INSTRUMENT")

  async def get_module_sap_serial_number(self) -> Optional[str]:
    """Gets the module SAP serial number."""
    return await self.send_command("?SAP_SERIAL_INSTRUMENT SAP_NR_INSTRUMENT")

  async def get_instrument_type(self) -> Optional[str]:
    """Gets the instrument type."""
    return await self.send_command("?INFO INSTRUMENT_TYPE")

  async def get_hardware_version(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the hardware version."""
    command = "?INFO HARDWARE_VERSION"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_user_defined(self) -> Optional[str]:
    """Gets the user-defined information."""
    return await self.send_command("?INFO USERDEFINED")

  async def get_available_modules(self) -> Optional[str]:
    """Gets the list of available modules."""
    return await self.send_command("#MODULE")

  async def get_expected_modules(self) -> Optional[str]:
    """Gets the list of expected modules."""
    return await self.send_command("#MODULE EXPECTED")

  async def get_expected_usb_modules(self) -> Optional[str]:
    """Gets the list of expected USB modules."""
    return await self.send_command("#MODULE EXPECTED_USB")

  async def get_available_sub_modules(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the list of available sub-modules."""
    command = "#MODULE SUB"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_expected_sub_modules(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the list of expected sub-modules."""
    command = "#MODULE EXPECTED_SUB"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_optional_modules(self) -> Optional[str]:
    """Gets the list of optional modules."""
    return await self.send_command("#MODULE DYNAMIC")

  async def get_available_functions(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the list of available functions."""
    command = "#FUNCTION"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    return await self.send_command(command)

  async def get_limit_value(
    self,
    name: str,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the integer limit value for a given name."""
    command = "?LIMIT"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    command += f" {name.upper()}"
    return await self.send_command(command)

  async def get_double_limit_value(
    self,
    name: str,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the double limit value for a given name."""
    command = "?LIMIT"
    if hw_module:
      command += f" MODULE={hw_module.value}"
    if number is not None:
      command += f" NUMBER={number}"
    if subcomponent:
      command += f" SUB={subcomponent}"
    command += f" {name.upper()}"
    return await self.send_command(command)

  async def get_available_message_types(self) -> Optional[str]:
    """Gets the available message types."""
    response = await self.send_command("#MESSAGE TYPE")
    return response

  async def get_interval_range(self, message_type: InstrumentMessageType) -> Optional[str]:
    """Gets the interval range for a message type."""
    return await self.send_command(f"#MESSAGE TYPE={message_type.value.upper()} TIME_INTERVAL")

  async def get_current_interval(self, message_type: InstrumentMessageType) -> Optional[str]:
    """Gets the current interval for a message type."""
    return await self.send_command(f"?MESSAGE TYPE={message_type.value.upper()} TIME_INTERVAL")

  async def set_interval(self, message_type: InstrumentMessageType, interval: int) -> Optional[str]:
    """Sets the interval for a message type."""
    return await self.send_command(
      f"MESSAGE TYPE={message_type.value.upper()} TIME_INTERVAL={interval}"
    )

  async def turn_all_interval_messages_off(self) -> Optional[str]:
    """Turns off all interval messages."""
    return await self.send_command("MESSAGE TYPE=ALL TIME_INTERVAL=0")

  def _create_target_string(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> str:
    target_string = ""
    if hw_module:
      target_string += f" MODULE={hw_module.value}"
    if number is not None:
      target_string += f" NUMBER={number}"
    if subcomponent:
      target_string += f" SUB={subcomponent}"
    return target_string

  async def get_upload_memory_scope(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the upload memory scope."""
    target_string = self._create_target_string(hw_module, number, subcomponent)
    response = await self.send_command(f"#UPLOAD TYPE{target_string}")
    return response

  async def get_ranges_for_upload_memory_scope(
    self,
    memory_scope: str,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the ranges for an upload memory scope."""
    target_string = self._create_target_string(hw_module, number, subcomponent)
    return await self.send_command(f"#UPLOAD{target_string} TYPE={memory_scope}")

  async def prepare_upload(
    self,
    memory_scope: str,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Prepares the device for upload."""
    target_string = self._create_target_string(hw_module, number, subcomponent)
    # This command uses data channel, which is not fully implemented yet.
    logging.warning("Prepare upload uses data channel, not fully implemented.")
    return await self.send_command(f"UPLOAD PREPARE{target_string} TYPE={memory_scope}")

  async def upload_block(
    self,
    offset: int,
    size: int,
    timeout: int,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Uploads a block of data."""
    target_string = self._create_target_string(hw_module, number, subcomponent)
    # This command uses data channel, which is not fully implemented yet.
    logging.warning("Upload block uses data channel, not fully implemented.")
    return await self.send_command(f"UPLOAD BLOCK{target_string} OFFSET={offset} SIZE={size}")

  async def get_upload_sections(
    self,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the available upload sections."""
    target_string = self._create_target_string(hw_module, number, subcomponent)
    response = await self.send_command(f"#UPLOAD SECTION{target_string} NAME")
    return response

  async def upload_section(
    self,
    section_name: str,
    size: int,
    timeout: int,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Uploads a specific section."""
    target_string = self._create_target_string(hw_module, number, subcomponent)
    # This command uses data channel, which is not fully implemented yet.
    logging.warning("Upload section uses data channel, not fully implemented.")
    return await self.send_command(f"UPLOAD SECTION{target_string} NAME={section_name} SIZE={size}")

  async def get_upload_section_size(
    self,
    section_name: str,
    hw_module: Optional[ModuleType] = None,
    number: Optional[int] = None,
    subcomponent: Optional[str] = None,
  ) -> Optional[str]:
    """Gets the size of a specific upload section."""
    target_string = self._create_target_string(hw_module, number, subcomponent)
    return await self.send_command(f"?UPLOAD SECTION{target_string} NAME={section_name} SIZE")



from pylabrobot.tilting import TilterBackend


class TilterChatterboxBackend(TilterBackend):
  async def setup(self):
    print("Setting up tilter.")

  async def stop(self):
    print("Stopping tilter.")

  async def set_angle(self, angle: float):
    print(f"Setting the angle to {angle}.")



from typing import Callable, Dict, Optional, cast

from .container import Container
from .coordinate import Coordinate
from .resource import Resource


class PetriDish(Container):
  """A petri dish"""

  def __init__(
    self,
    name: str,
    diameter: float,
    height: float,
    material_z_thickness: Optional[float] = None,
    category: str = "petri_dish",
    model: Optional[str] = None,
    max_volume: Optional[float] = None,
    compute_volume_from_height: Optional[Callable[[float], float]] = None,
    compute_height_from_volume: Optional[Callable[[float], float]] = None,
    height_volume_data: Optional[Dict[float, float]] = None,
  ):
    super().__init__(
      name=name,
      size_x=diameter,
      size_y=diameter,
      size_z=height,
      material_z_thickness=material_z_thickness,
      category=category,
      model=model,
      max_volume=max_volume,
      compute_volume_from_height=compute_volume_from_height,
      compute_height_from_volume=compute_height_from_volume,
      height_volume_data=height_volume_data,
    )
    self.diameter = diameter
    self.height = height

  def serialize(self):
    super_serialized = super().serialize()
    for key in ["size_x", "size_y", "size_z"]:
      del super_serialized[key]

    return {
      **super_serialized,
      "diameter": self.diameter,
      "height": self.height,
    }


class PetriDishHolder(Resource):
  """3d printed holder for petri dish, size of a 96 well plate"""

  def __init__(
    self,
    name: str,
    size_x: float = 127.76,
    size_y: float = 85.48,
    size_z: float = 14.5,
    category: str = "petri_dish_holder",
    model: Optional[str] = None,
  ):
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      category=category,
      model=model,
    )

  def assign_child_resource(
    self,
    resource: Resource,
    location: Optional[Coordinate],
    reassign: bool = True,
  ):
    """Can only assign a single PetriDish"""
    if not isinstance(resource, PetriDish):
      raise TypeError("Can only assign PetriDish to PetriDishHolder")

    if len(self.children) > 0:
      raise ValueError("Can only assign a single PetriDish to PetriDishHolder")

    super().assign_child_resource(resource, location, reassign)

  @property
  def dish(self) -> Optional[PetriDish]:
    if len(self.children) == 0:
      return None
    return cast(PetriDish, self.children[0])



import unittest

from pylabrobot.resources import Coordinate

from .planning import group_by_x_batch_by_xy


class TestGroupByXBatchByXY(unittest.TestCase):
  """Tests for group_by_x_batch_by_xy."""

  def test_single_location(self):
    locations = [Coordinate(100.0, 200.0, 0)]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0])
    self.assertEqual(result, {100.0: [[0]]})

  def test_same_x_different_y_fits_in_one_batch(self):
    locations = [
      Coordinate(100.0, 200.0, 0),
      Coordinate(100.0, 180.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 1])
    # y diff = 20 >= 9*1, fits in one batch
    self.assertEqual(result, {100.0: [[0, 1]]})

  def test_same_x_too_close_y_splits_batches(self):
    locations = [
      Coordinate(100.0, 200.0, 0),
      Coordinate(100.0, 195.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 1])
    # y diff = 5 < 9*1, separate batches
    self.assertEqual(result, {100.0: [[0], [1]]})

  def test_different_x_groups(self):
    locations = [
      Coordinate(100.0, 200.0, 0),
      Coordinate(200.0, 200.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 1])
    self.assertEqual(result, {100.0: [[0]], 200.0: [[1]]})

  def test_x_rounding(self):
    locations = [
      Coordinate(100.04, 200.0, 0),
      Coordinate(100.02, 180.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 1])
    # Both round to 100.0
    self.assertEqual(result, {100.0: [[0, 1]]})

  def test_two_channels_same_x(self):
    locations = [Coordinate(100.0, 200.0, 0), Coordinate(100.0, 180.0, 0)]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 1])
    self.assertEqual(result, {100.0: [[0, 1]]})

  def test_empty_use_channels_raises(self):
    with self.assertRaises(ValueError):
      group_by_x_batch_by_xy(
        locations=[Coordinate(100.0, 200.0, 0)],
        use_channels=[],
      )

  def test_non_adjacent_channels(self):
    locations = [
      Coordinate(100.0, 300.0, 0),
      Coordinate(100.0, 200.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 5])
    # y diff = 100 >= 9*(5-0) = 45, fits in one batch
    self.assertEqual(result, {100.0: [[0, 1]]})

  def test_non_adjacent_channels_too_close(self):
    locations = [
      Coordinate(100.0, 240.0, 0),
      Coordinate(100.0, 200.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 5])
    # y diff = 40 < 9*(5-0) = 45, separate batches
    self.assertEqual(result, {100.0: [[0], [1]]})

  def test_sorted_by_x(self):
    locations = [
      Coordinate(300.0, 200.0, 0),
      Coordinate(100.0, 200.0, 0),
      Coordinate(200.0, 200.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 1, 2])
    self.assertEqual(result, {100.0: [[1]], 200.0: [[2]], 300.0: [[0]]})

  def test_multiple_batches_in_one_x_group(self):
    locations = [
      Coordinate(100.0, 200.0, 0),
      Coordinate(100.0, 197.0, 0),
      Coordinate(100.0, 194.0, 0),
      Coordinate(100.0, 191.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 1, 2, 3])
    # Each consecutive pair has y diff = 3 < 9, so each in its own batch
    self.assertEqual(result, {100.0: [[0], [1], [2], [3]]})

  def test_duplicate_channels_split_into_separate_batches(self):
    locations = [
      Coordinate(100.0, 200.0, 0),
      Coordinate(100.0, 180.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 0])
    self.assertEqual(result, {100.0: [[0], [1]]})

  def test_duplicate_channels_three_ops(self):
    locations = [
      Coordinate(100.0, 200.0, 0),
      Coordinate(100.0, 180.0, 0),
      Coordinate(100.0, 160.0, 0),
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[0, 0, 0])
    self.assertEqual(result, {100.0: [[0], [1], [2]]})

  def test_channels_sorted_by_channel_index_within_x_group(self):
    locations = [
      Coordinate(100.0, 180.0, 0),  # channel 2
      Coordinate(100.0, 200.0, 0),  # channel 0
    ]
    result = group_by_x_batch_by_xy(locations=locations, use_channels=[2, 0])
    # Channel 0 (index 1) sorted before channel 2 (index 0)
    self.assertEqual(result, {100.0: [[1, 0]]})


if __name__ == "__main__":
  unittest.main()



from typing import Optional, cast

from pylabrobot.resources.carrier import Carrier
from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.resources.deck import Deck
from pylabrobot.resources.plate import Plate
from pylabrobot.resources.resource import Resource
from pylabrobot.resources.tecan.tecan_resource import TecanResource
from pylabrobot.resources.tecan.wash import (
  Wash_Station,
  Wash_Station_Cleaner_deep,
  Wash_Station_Cleaner_shallow,
  Wash_Station_Waste,
)
from pylabrobot.resources.tip_rack import TipRack
from pylabrobot.resources.trash import Trash

_RAILS_WIDTH = 25

EVO100_NUM_RAILS = 30
EVO100_SIZE_X = 940
EVO100_SIZE_Y = 780
EVO100_SIZE_Z = 765

EVO150_NUM_RAILS = 45
EVO150_SIZE_X = 1315
EVO150_SIZE_Y = 780
EVO150_SIZE_Z = 765

EVO200_NUM_RAILS = 69
EVO200_SIZE_X = 1915
EVO200_SIZE_Y = 780
EVO200_SIZE_Z = 765


class TecanDeck(Deck):
  """Tecan decks"""

  def __init__(
    self,
    num_rails: int,
    size_x: float,
    size_y: float,
    size_z: float,
    name: str = "deck",
    category: str = "deck",
    origin: Coordinate = Coordinate(0, 0, 0),
    with_wash_station: bool = True,
  ):
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      category=category,
      origin=origin,
    )
    self.num_rails = num_rails

    if with_wash_station:
      wash = Wash_Station(name="wash_station")
      wash[0] = Wash_Station_Cleaner_deep(name="wash_clean_deep")
      wash[1] = Wash_Station_Waste(name="wash_waste")
      wash[2] = Wash_Station_Cleaner_shallow(name="wash_clean_shallow")
      self.assign_child_resource(wash, rails=1)

  def serialize(self) -> dict:
    return {**super().serialize(), "num_rails": self.num_rails}

  def assign_child_resource(
    self,
    resource: Resource,
    location: Optional[Coordinate] = None,
    reassign: bool = False,
    rails: Optional[int] = None,
    replace=False,
  ):
    """Assign a new deck resource."""

    if rails is not None and not 1 <= rails <= self.num_rails:
      raise ValueError(f"Rails must be between 1 and {self.num_rails}.")

    # Check if resource exists.
    if self.has_resource(resource.name):
      if replace:
        # unassign first, so we don't have problems with location checking later.
        cast(Resource, self.get_resource(resource.name)).unassign()
      else:
        raise ValueError(f"Resource with name '{resource.name}' already defined.")

    if rails is not None:
      resource_location = self._coordinate_for_rails(rails, resource)
    elif location is not None:
      resource_location = location
    else:
      resource_location = None  # unknown resource location

    if resource_location is not None:
      if (
        resource_location.x + resource.get_absolute_size_x() > self.get_absolute_size_x()
        and rails is not None
      ):
        raise ValueError(
          f"Resource with width {resource.get_absolute_size_x()} does not fit at rails {rails}."
        )

      # Check if there is space for this new resource.
      for og_resource in self.children:
        og_x = cast(Coordinate, og_resource.location).x
        og_y = cast(Coordinate, og_resource.location).y

        # A resource is not allowed to overlap with another resource. Resources overlap when a
        # corner of one resource is inside the boundaries other resource.
        if (
          og_x <= resource_location.x < og_x + og_resource.get_absolute_size_x()
          or og_x
          <= resource_location.x + resource.get_absolute_size_x()
          < og_x + og_resource.get_absolute_size_x()
        ) and (
          og_y <= resource_location.y < og_y + og_resource.get_absolute_size_y()
          or og_y
          <= resource_location.y + resource.get_absolute_size_y()
          < og_y + og_resource.get_absolute_size_y()
        ):
          raise ValueError(
            f"Location {resource_location} is already occupied by resource '{og_resource.name}'."
          )

    return super().assign_child_resource(resource, location=resource_location)

  def _coordinate_for_rails(self, rails: int, resource: Resource):
    """Convert a rail identifier and resource to a location."""

    if not isinstance(resource, TecanResource):
      raise ValueError(f"Resource {resource} is not a Tecan resource.")

    return Coordinate(
      (rails - 1) * _RAILS_WIDTH - resource.off_x + 100,
      resource.off_y + 345 - resource.get_absolute_size_y(),
      0,
    )  # TODO: verify

  def _rails_for_x_coordinate(self, x: float):
    """Convert an x coordinate to a rail identifier."""

    return round((x + _RAILS_WIDTH - 101) / _RAILS_WIDTH) + 1

  def summary(self) -> str:
    """Return a summary of the deck."""

    if len(self.get_all_resources()) == 0:
      raise ValueError(
        "This liquid editor does not have any resources yet. "
        "Build a layout first by calling `assign_child_resource()`. "
      )

    # Print header.
    summary_ = "Rail" + " " * 5 + "Resource" + " " * 19 + "Type" + " " * 16 + "Coordinates (mm)\n"
    summary_ += "=" * 95 + "\n"

    def parse_resource(resource) -> str:
      # TODO: print something else if resource is not assigned to a rails.
      rails = self._rails_for_x_coordinate(resource.location.x)
      rail_label = f"({rails})" if rails is not None else "     "
      r_summary = (
        f"{rail_label:4} ├── {resource.name:27}"
        + f"{resource.__class__.__name__:20}"
        + f"{resource.get_location_wrt(self)}\n"
      )

      if isinstance(resource, Carrier):
        for site in resource.sites.values():
          if site.resource is None:
            r_summary += "     │   ├── <empty>\n"
          else:
            subresource = site.resource
            if isinstance(subresource, (TipRack, Plate)):
              location = (
                subresource.get_item("A1").get_location_wrt(self)
                + subresource.get_item("A1").center()
              )
            else:
              location = subresource.get_location_wrt(self)
            r_summary += (
              f"     │   ├── {subresource.name:23}"
              + f"{subresource.__class__.__name__:20}"
              + f"{location}\n"
            )

      return r_summary

    # Sort resources by rails, left to right in reality.
    sorted_resources = sorted(self.children, key=lambda r: r.get_location_wrt(self).x)

    # Print table body.
    summary_ += parse_resource(sorted_resources[0])
    for resource in sorted_resources[1:]:
      summary_ += "     │\n"
      summary_ += parse_resource(resource)

    return summary_

  def clear(self, include_trash: bool = False):
    """Clear the deck, removing all resources except the trash areas."""
    children_names = [child.name for child in self.children]
    for resource_name in children_names:
      resource = self.get_resource(resource_name)
      if isinstance(resource, Trash) and not include_trash:
        continue
      if resource.name == "wash_station":
        continue
      resource.unassign()


def EVO100Deck(origin: Coordinate = Coordinate(0, 0, 0)) -> TecanDeck:
  """EVO100 deck.

  Sizes from operating manual
  """

  return TecanDeck(
    num_rails=EVO100_NUM_RAILS,
    size_x=EVO100_SIZE_X,
    size_y=EVO100_SIZE_Y,
    size_z=EVO100_SIZE_Z,
    origin=origin,
  )


def EVO150Deck(origin: Coordinate = Coordinate(0, 0, 0)) -> TecanDeck:
  """EVO150 deck.

  Sizes from operating manual
  """

  return TecanDeck(
    num_rails=EVO150_NUM_RAILS,
    size_x=EVO150_SIZE_X,
    size_y=EVO150_SIZE_Y,
    size_z=EVO150_SIZE_Z,
    origin=origin,
  )


def EVO200Deck(origin: Coordinate = Coordinate(0, 0, 0)) -> TecanDeck:
  """EVO200 deck.

  Sizes from operating manual
  """

  return TecanDeck(
    num_rails=EVO200_NUM_RAILS,
    size_x=EVO200_SIZE_X,
    size_y=EVO200_SIZE_Y,
    size_z=EVO200_SIZE_Z,
    origin=origin,
  )



async def distribute_fixed_volume_w_dispense_offsets(
  liquid_handler,
  source_container_name: str,  # a single container
  vol: float,
  destination_plate_name: str,
  destination_wells: List[str],  # List of well names: ["A1", "B2", ...]
  destination_offsets: List[Coordinate],  # List of offsets matching the wells
  use_channels: List[int],
  hlc: HamiltonLiquidClass | None,
  # Defaults
  aspiration_surface_following_distance: float = 4.0,
  dispense_surface_following_distance: float = 0.0,
  aspiration_lld_mode: Literal["cLLD", "pLLD"] = "cLLD",
  tip_safety_margin: float = 0.95,
  verbose: bool = True,
):
  """
  Distribute a fixed volume to multiple destination wells with custom offsets per well.

  Args:
      destination_wells: List of well identifiers (e.g., ["A1", "B2", "C3"])
      destination_offsets: List of Coordinate objects, one per well (must match length of destination_wells)

  Optimization improvements:
  - Separated wells and offsets into two lists for cleaner API
  - Cached repeated calculations (location lookups, constants)
  - Pre-calculate constants outside loops
  - Uses spatial chunking for efficient pipetting patterns
  """

  assert isinstance(use_channels, list) and 1 <= len(use_channels) <= 8, (
    "Number of use_channels must be between 1 and 8."
  )

  assert len(destination_wells) == len(destination_offsets), (
    "destination_wells and destination_offsets must have the same length."
  )

  # Pre-calculate volumes and limits
  asp_volume_per_dispense = hlc.compute_corrected_volume(vol)
  max_tip_volume = (
    min([ch.get_tip().maximal_volume for ch in liquid_handler.head.values() if ch.has_tip])
    * tip_safety_margin
  )
  dispenses_per_cycle = floor(max_tip_volume / asp_volume_per_dispense)
  num_channels = len(use_channels)

  # Get source and destination resources
  source_container = liquid_handler.deck.get_resource(source_container_name)
  dest_plate = liquid_handler.deck.get_resource(destination_plate_name)

  # Cache source container minimum height (computed once, used many times)
  source_min_height = source_container.get_absolute_location(z="cavity_bottom").z

  # Pre-calculate LLD mode (constant lookup moved outside loop)
  lld_mode_dict = {"fixed_z": 0, "cLLD": 1, "pLLD": 2}
  asp_lld_mode = liquid_handler.backend.LLDMode(lld_mode_dict[aspiration_lld_mode])
  disp_lld_mode = liquid_handler.backend.LLDMode(0)

  # Get well objects from plate and create mapping to offsets
  well_objects = dest_plate[destination_wells]
  well_to_offset = {well: offset for well, offset in zip(well_objects, destination_offsets)}

  # Use spatial chunking (groups by x-coordinate proximity, then sorts by y)
  chunked_wells = chunk_resource_list_by_x_then_sort(well_objects, max_chunk_size=num_channels)

  # Rebuild the chunks with their corresponding offsets
  chunked_wells_and_offsets = [
    [(well, well_to_offset[well]) for well in chunk] for chunk in chunked_wells
  ]

  # Group chunks into transfer cycles
  transfer_cycle_pattern = chunk_list(chunked_wells_and_offsets, chunk_size=dispenses_per_cycle)

  for asp_idx, single_cycle_list in enumerate(
    tqdm(transfer_cycle_pattern, desc="Aspiration cycle", disable=not verbose)
  ):
    max_no_channels = max(len(op) for op in single_cycle_list)
    no_dispenses = len(single_cycle_list)
    aspiration_vol = asp_volume_per_dispense * no_dispenses

    # Determine y_offset based on odd/even channels
    y_offset = 5.5 if max_no_channels % 2 != 0 else 0

    await liquid_handler.aspirate(
      resources=[source_container] * max_no_channels,
      vols=[aspiration_vol] * max_no_channels,
      use_channels=use_channels[:max_no_channels],
      hamilton_liquid_classes=[hlc] * max_no_channels,
      offsets=[Coordinate(x=0, y=y_offset, z=0)] * max_no_channels,
      lld_mode=[asp_lld_mode] * max_no_channels,
      immersion_depth_direction=[0] * max_no_channels,
      immersion_depth=[2] * max_no_channels,
      surface_following_distance=[aspiration_surface_following_distance] * max_no_channels,
      minimum_height=[source_min_height] * max_no_channels,
    )

    for disp_idx, wells_and_offsets_list in enumerate(
      tqdm(single_cycle_list, desc="Dispense step", disable=not verbose)
    ):
      # Unpack wells and offsets using zip (more pythonic)
      dest_well_list, dispense_offset_coordinate_list = zip(*wells_and_offsets_list)
      dest_well_list = list(dest_well_list)
      dispense_offset_coordinate_list = list(dispense_offset_coordinate_list)

      no_channels = len(dest_well_list)

      # Cache location lookups (avoid repeated method calls)
      first_well_top_z = dest_well_list[0].get_absolute_location(z="top").z
      first_well_bottom_z = dest_well_list[0].get_absolute_location(z="cavity_bottom").z
      traverse_height = first_well_top_z + 5

      await liquid_handler.dispense(
        resources=dest_well_list,
        vols=[vol] * no_channels,
        use_channels=use_channels[:no_channels],
        hamilton_liquid_classes=[hlc] * no_channels,
        lld_mode=[disp_lld_mode] * no_channels,
        offsets=dispense_offset_coordinate_list,
        surface_following_distance=[dispense_surface_following_distance] * no_channels,
        minimum_height=[first_well_bottom_z] * no_channels,
        minimum_traverse_height_at_beginning_of_a_command=traverse_height,
        min_z_endpos=traverse_height,
      )

      if verbose:
        print(f"{asp_idx=} | {disp_idx=} dispense: {[x.get_identifier() for x in dest_well_list]}")

    await liquid_handler.backend.move_all_channels_in_z_safety()

    empty_z_height = source_container.get_absolute_location(z="top").z - 2

    await liquid_handler.aspirate(
      resources=[source_container] * max_no_channels,
      vols=[0] * max_no_channels,
      use_channels=use_channels[:max_no_channels],
      hamilton_liquid_classes=[hlc] * max_no_channels,
      offsets=[Coordinate(x=0, y=y_offset, z=source_container.get_size_z())] * max_no_channels,
      lld_mode=[liquid_handler.backend.LLDMode(0)] * max_no_channels,
      immersion_depth_direction=[0] * max_no_channels,
      immersion_depth=[0] * max_no_channels,
      surface_following_distance=[0] * max_no_channels,
      pull_out_distance_transport_air=[0] * max_no_channels,
      minimum_height=[empty_z_height] * max_no_channels,
      min_z_endpos=empty_z_height,
    )

    await liquid_handler.backend.empty_channels()

    await liquid_handler.backend.move_all_channels_in_z_safety()



from pylabrobot.scales.scale_backend import ScaleBackend


class ScaleChatterboxBackend(ScaleBackend):
  """Chatter box backend for device-free testing. Prints out all operations."""

  def __init__(self, dummy_weight: float = 0.0) -> None:
    self._dummy_weight = dummy_weight

  async def setup(self) -> None:
    print("Setting up the scale.")

  async def stop(self) -> None:
    print("Stopping the scale.")

  async def tare(self):
    print("Taring the scale")

  async def get_weight(self) -> float:
    print("Getting the weight")
    return self._dummy_weight

  async def zero(self):
    print("Zeroing the scale")



import enum
from typing import Optional

from pylabrobot.resources.tip import Tip


class TipType(enum.Enum):
  """Tip type"""

  STANDARD = "Standard"
  DITI = "DiTi"
  STDLOWVOL = "StandardLowVolume"
  DITILOWVOL = "DiTiLowVolume"
  MCADITI = "MCADiTi"
  AIRDITI = "ZaapDiTi"


class TecanTip(Tip):
  """Represents a single tip for Tecan instruments."""

  def __init__(
    self,
    has_filter: bool,
    total_tip_length: float,
    maximal_volume: float,
    tip_type: TipType,
    fitting_depth: float = 0,
    name: Optional[str] = None,
  ):
    super().__init__(
      has_filter=has_filter,
      total_tip_length=total_tip_length,
      maximal_volume=maximal_volume,
      fitting_depth=fitting_depth,
      name=name,
    )
    self.tip_type = tip_type


def standard_fixed_tip(name: Optional[str] = None) -> TecanTip:
  """Default standard fixed tip"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=39.0,
    maximal_volume=1000,
    tip_type=TipType.STANDARD,
  )


def DiTi_100ul_Te_MO_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_100ul_Te_MO"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=42.0,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_Te_MO_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_Te_MO"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=28.0,
    maximal_volume=60.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_Te_MO_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_Te_MO"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=42.0,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def DiTi_100ul_Filter_Te_MO_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_100ul_Filter_Te_MO"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=42.0,
    maximal_volume=90.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_Filter_Te_MO_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_Filter_Te_MO"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=42.0,
    maximal_volume=170.0,
    tip_type=TipType.DITI,
  )


def Adapter_96_DiTi_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for Adapter_96_DiTi_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=15.9,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def Adapter_DiTi_Combo_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for Adapter_DiTi_Combo_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=15.9,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def Adapter_DiTi_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for Adapter_DiTi_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=15.9,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def DiTi_100ul_Filter_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_100ul_Filter_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=90.0,
    tip_type=TipType.DITI,
  )


def DiTi_100ul_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_100ul_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_Filter_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_Filter_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=170.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=29.6,
    maximal_volume=60.0,
    tip_type=TipType.DITI,
  )


def Base_Nested_DiTi_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for Base_Nested_DiTi_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def DiTi_100ul_Nested_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_100ul_Nested_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def DiTi_100ul_SBS_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_100ul_SBS_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_SBS_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_SBS_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_SBS_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_SBS_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=29.6,
    maximal_volume=60.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_Nested_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_Nested_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=29.6,
    maximal_volume=60.0,
    tip_type=TipType.DITI,
  )


def Adapter_96_DiTi_1to1_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for Adapter_96_DiTi_1to1_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=25.2,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_Nested_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_Nested_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_w_b_filter_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_w_b_filter_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=175.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_wide_bore_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_wide_bore_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=43.1,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def Adapter_96_DiTi_4to1_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for Adapter_96_DiTi_4to1_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=25.2,
    maximal_volume=110.0,
    tip_type=TipType.DITI,
  )


def DiTi_500ul_Filter_SBS_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_500ul_Filter_SBS_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=54.0,
    maximal_volume=405.0,
    tip_type=TipType.DITI,
  )


def DiTi_500ul_SBS_MCA96_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_500ul_SBS_MCA96"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=54.0,
    maximal_volume=502.0,
    tip_type=TipType.DITI,
  )


def DiTi_Nested_Waste_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_Nested_Waste_MCA384"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=0.0,
    maximal_volume=0.0,
    tip_type=TipType.DITI,
  )


def DiTi_1000ul_SBS_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_1000ul_SBS_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=32.6,
    maximal_volume=1100.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_SBS_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_SBS_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=21.0,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_SBS_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_SBS_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.3,
    maximal_volume=55.0,
    tip_type=TipType.DITI,
  )


def DiTi_5000ul_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_5000ul_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=46.6,
    maximal_volume=5130.0,
    tip_type=TipType.DITI,
  )


def DiTi_5000ul_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_5000ul_Filter_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=46.6,
    maximal_volume=5130.0,
    tip_type=TipType.DITI,
  )


def DiTi_125ul_Filter_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_125ul_Filter_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=47.3,
    maximal_volume=115.0,
    tip_type=TipType.DITI,
  )


def DiTi_125ul_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_125ul_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=47.3,
    maximal_volume=129.0,
    tip_type=TipType.DITI,
  )


def DiTi_15ul_Filter_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_15ul_Filter_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=28.6,
    maximal_volume=18.0,
    tip_type=TipType.DITI,
  )


def DiTi_15ul_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_15ul_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=28.6,
    maximal_volume=19.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_Filter_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_Filter_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=41.4,
    maximal_volume=44.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_MCA384_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_MCA384"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=41.4,
    maximal_volume=53.0,
    tip_type=TipType.DITI,
  )


def DiTi_1000ul_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_1000ul_Filter_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=32.6,
    maximal_volume=1050.0,
    tip_type=TipType.DITI,
  )


def DiTi_1000ul_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_1000ul_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=32.6,
    maximal_volume=1100.0,
    tip_type=TipType.DITI,
  )


def DiTi_10ul_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_10ul_Filter_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-31.3,
    maximal_volume=12.0,
    tip_type=TipType.DITI,
  )


def DiTi_10ul_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_10ul_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-31.3,
    maximal_volume=23.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_Filter_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.0,
    maximal_volume=210.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.0,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_Filter_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.3,
    maximal_volume=55.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.3,
    maximal_volume=55.0,
    tip_type=TipType.DITI,
  )


def DiTi_350ul_Nested_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_350ul_Nested_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.0,
    maximal_volume=390.0,
    tip_type=TipType.DITI,
  )


def DiTi_10ul_Filter_LiHa_L_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_10ul_Filter_LiHa_L"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-28.1,
    maximal_volume=12.0,
    tip_type=TipType.DITI,
  )


def DiTi_10ul_Filter_Nested_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_10ul_Filter_Nested_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-28.1,
    maximal_volume=12.0,
    tip_type=TipType.DITI,
  )


def DiTi_10ul_LiHa_L_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_10ul_LiHa_L"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-28.1,
    maximal_volume=23.0,
    tip_type=TipType.DITI,
  )


def DiTi_10ul_Nested_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_10ul_Nested_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-28.1,
    maximal_volume=23.0,
    tip_type=TipType.DITI,
  )


def DiTi_10ul_SBS_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_10ul_SBS_Filter_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-28.1,
    maximal_volume=12.0,
    tip_type=TipType.DITI,
  )


def DiTi_10ul_SBS_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_10ul_SBS_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-28.1,
    maximal_volume=23.0,
    tip_type=TipType.DITI,
  )


def DiTi_1000ul_W_B_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_1000ul_W_B_Filter_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=32.0,
    maximal_volume=995.0,
    tip_type=TipType.DITI,
  )


def DiTi_1000ul_CL_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_1000ul_CL_Filter_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=32.6,
    maximal_volume=1050.0,
    tip_type=TipType.DITI,
  )


def DiTi_1000ul_CL_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_1000ul_CL_LiHa"""
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=32.6,
    maximal_volume=1100.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_CL_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_CL_Filter_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.0,
    maximal_volume=210.0,
    tip_type=TipType.DITI,
  )


def DiTi_200ul_CL_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_200ul_CL_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.0,
    maximal_volume=220.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_CL_Filter_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_CL_Filter_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.3,
    maximal_volume=55.0,
    tip_type=TipType.DITI,
  )


def DiTi_50ul_CL_LiHa_tip(name: Optional[str] = None) -> TecanTip:
  """Tip for DiTi_50ul_CL_LiHa"""
  print("WARNING: total_tip_length <= 0.")
  print("Please get in touch at https://discuss.pylabrobot.org")
  return TecanTip(
    name=name,
    has_filter=False,
    total_tip_length=-5.3,
    maximal_volume=60.0,
    tip_type=TipType.DITI,
  )



# mypy: disable-error-code="union-attr,assignment,arg-type"
"""Tests for BioTek EL406 plate washer backend - Shake operations.

This module contains tests for shake-related step methods:
- shake (SHAKE_SOAK)
"""

import unittest

from pylabrobot.plate_washing.biotek.el406 import ExperimentalBioTekEL406Backend
from pylabrobot.plate_washing.biotek.el406.mock_tests import PT96, EL406TestCase


class TestEL406BackendShake(EL406TestCase):
  """Test EL406 shake functionality."""

  async def test_shake_sends_command(self):
    """Shake should send correct command."""
    initial_count = len(self.backend.io.written_data)
    await self.backend.shake(PT96, duration=10, intensity="Medium")

    self.assertGreater(len(self.backend.io.written_data), initial_count)

  async def test_shake_validates_intensity(self):
    """Shake should validate intensity value."""
    with self.assertRaises(ValueError):
      await self.backend.shake(PT96, duration=10, intensity="invalid")

  async def test_shake_validates_both_zero(self):
    """Shake should raise ValueError when both duration and soak_duration are 0."""
    with self.assertRaises(ValueError):
      await self.backend.shake(PT96, duration=0, soak_duration=0)

  async def test_shake_validates_negative_duration(self):
    """Shake should raise ValueError for negative duration."""
    with self.assertRaises(ValueError) as ctx:
      await self.backend.shake(PT96, duration=-5)

    self.assertIn("duration", str(ctx.exception).lower())
    self.assertIn("-5", str(ctx.exception))

  async def test_shake_validates_negative_soak(self):
    """Shake should raise ValueError for negative soak_duration."""
    with self.assertRaises(ValueError):
      await self.backend.shake(PT96, duration=10, soak_duration=-1)

  async def test_shake_validates_duration_exceeds_max(self):
    """Shake should raise ValueError when duration exceeds 3599s (59:59)."""
    with self.assertRaises(ValueError):
      await self.backend.shake(PT96, duration=3600)

  async def test_shake_validates_soak_exceeds_max(self):
    """Shake should raise ValueError when soak_duration exceeds 3599s (59:59)."""
    with self.assertRaises(ValueError):
      await self.backend.shake(PT96, duration=10, soak_duration=3600)

  async def test_shake_soak_only(self):
    """Shake with duration=0 and soak_duration>0 should work (soak only)."""
    initial_count = len(self.backend.io.written_data)
    await self.backend.shake(PT96, duration=0, soak_duration=10)

    self.assertGreater(len(self.backend.io.written_data), initial_count)


class TestShakeCommandEncoding(unittest.TestCase):
  """Test shake command binary encoding."""

  def setUp(self):
    self.backend = ExperimentalBioTekEL406Backend()

  def test_shake_command_basic(self):
    """Basic shake: 10 seconds, medium intensity."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=10.0,
      soak_duration=0.0,
      intensity="Medium",
      shake_enabled=True,
    )

    self.assertEqual(cmd[0], 0x04)
    self.assertEqual(cmd[1], 0x01)
    self.assertEqual(cmd[2], 0x0A)
    self.assertEqual(cmd[3], 0x00)
    self.assertEqual(cmd[4], 0x03)
    self.assertEqual(cmd[5], 0x00)
    self.assertEqual(cmd[6], 0x00)
    self.assertEqual(cmd[7], 0x00)
    self.assertEqual(cmd[8:12], bytes([0, 0, 0, 0]))

  def test_shake_command_variable_intensity(self):
    """Variable intensity encoding."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=30.0,
      soak_duration=0.0,
      intensity="Variable",
      shake_enabled=True,
    )

    self.assertEqual(cmd[4], 0x01)

  def test_shake_command_encoding_durations(self):
    """Verify encoding for various shake durations."""
    cases = [
      (30.0, "04011e000300000000000000"),  # 00:30
      (60.0, "04013c000300000000000000"),  # 01:00
      (300.0, "04012c010300000000000000"),  # 05:00
    ]
    for duration, expected_hex in cases:
      with self.subTest(duration=duration):
        cmd = self.backend._build_shake_command(
          PT96,
          shake_duration=duration,
          soak_duration=0.0,
          intensity="Medium",
          shake_enabled=True,
          move_home_first=True,
        )
        self.assertEqual(cmd, bytes.fromhex(expected_hex))

  def test_shake_command_encoding_shake_disabled(self):
    """Shake disabled should zero the duration."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=30.0,
      soak_duration=0.0,
      intensity="Medium",
      shake_enabled=False,
      move_home_first=True,
    )

    expected = bytes.fromhex("040100000300000000000000")
    self.assertEqual(cmd, expected)

  def test_shake_command_encoding_move_home_false(self):
    """Verify encoding with move_home_first=false."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=30.0,
      soak_duration=0.0,
      intensity="Medium",
      shake_enabled=True,
      move_home_first=False,
    )

    expected = bytes.fromhex("04001e000300000000000000")
    self.assertEqual(cmd, expected)

  def test_shake_command_encoding_soak_30s(self):
    """Verify encoding with 30s soak duration."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=30.0,
      soak_duration=30.0,
      intensity="Medium",
      shake_enabled=True,
      move_home_first=True,
    )

    expected = bytes.fromhex("04011e0003001e0000000000")
    self.assertEqual(cmd, expected)

  def test_shake_command_encoding_soak_60s(self):
    """Verify encoding with 60s soak duration."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=30.0,
      soak_duration=60.0,
      intensity="Medium",
      shake_enabled=True,
      move_home_first=True,
    )

    expected = bytes.fromhex("04011e0003003c0000000000")
    self.assertEqual(cmd, expected)

  def test_shake_command_encoding_slow_frequency(self):
    """Verify encoding with slow intensity."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=30.0,
      soak_duration=0.0,
      intensity="Slow",
      shake_enabled=True,
      move_home_first=True,
    )

    expected = bytes.fromhex("04011e000200000000000000")
    self.assertEqual(cmd, expected)

  def test_shake_command_encoding_fast_frequency(self):
    """Verify encoding with fast intensity."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=30.0,
      soak_duration=0.0,
      intensity="Fast",
      shake_enabled=True,
      move_home_first=True,
    )

    expected = bytes.fromhex("04011e000400000000000000")
    self.assertEqual(cmd, expected)

  def test_shake_command_encoding_complex(self):
    """Verify encoding with combined shake, soak, and slow intensity."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=300.0,
      soak_duration=120.0,
      intensity="Slow",
      shake_enabled=True,
      move_home_first=True,
    )

    expected = bytes.fromhex("04012c010200780000000000")
    self.assertEqual(cmd, expected)

  def test_shake_command_encoding_move_home_false_with_soak(self):
    """Verify encoding with move_home_first=false and soak."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=30.0,
      soak_duration=60.0,
      intensity="Medium",
      shake_enabled=True,
      move_home_first=False,
    )

    expected = bytes.fromhex("04001e0003003c0000000000")
    self.assertEqual(cmd, expected)

  def test_shake_command_max_duration_encoding(self):
    """Verify encoding with maximum duration (3599s)."""
    cmd = self.backend._build_shake_command(
      PT96,
      shake_duration=3599,
      soak_duration=3599,
      intensity="Medium",
      shake_enabled=True,
      move_home_first=True,
    )

    expected = bytes.fromhex("04010f0e03000f0e00000000")
    self.assertEqual(cmd, expected)


if __name__ == "__main__":
  unittest.main()



import warnings

# implemented as PlateAdapter to enable simple and fast assignment
# of plates to them, with self-correcting location placement
from pylabrobot.resources.plate_adapter import PlateAdapter


def alpaqua_96_plateadapter_magnum_flx(name: str) -> PlateAdapter:
  """Alpaqua Engineering LLC cat. no.: A000400
  Magnetic rack for 96-well plates.
  implemented as PlateAdapter to enable simple and fast assignment of
    plates to them, with self-correcting location placement
  """
  return PlateAdapter(
    name=name,
    size_x=127.76,
    size_y=85.48,
    size_z=35.0,
    dx=9.8,
    dy=6.8,
    dz=27.5,  # refers to magnet hole bottom
    plate_z_offset=0.0,  # adjust at runtime based on plate's well geometry
    adapter_hole_size_x=8.0,
    adapter_hole_size_y=8.0,
    adapter_hole_size_z=8.0,  # guesstimate
    model=alpaqua_96_plateadapter_magnum_flx.__name__,
  )


# Deprecated names for backwards compatibility
# TODO: Remove >2026-02


def Alpaqua_96_magnum_flx(name: str) -> PlateAdapter:
  """Deprecated alias for `alpaqua_96_plateadapter_magnum_flx`."""
  warnings.warn(
    "Alpaqua_96_magnum_flx is deprecated. Use 'alpaqua_96_plateadapter_magnum_flx' instead.",
    DeprecationWarning,
    stacklevel=2,
  )
  return alpaqua_96_plateadapter_magnum_flx(name)



"""Primitive byte de/serialization. Nice wrapper around struct packing/unpacking.

This module provides low-level byte serialization/deserialization without any
protocol-specific wrapping.

Example:
  # Writing
  data = Writer().u8(1).u16(100).string("test").finish()

  # Reading
  reader = Reader(data)
  val1 = reader.u8()
  val2 = reader.u16()
  val3 = reader.string()
"""

from __future__ import annotations

import struct
from io import BytesIO


class Writer:
  """Raw byte writer.

  Provides fluent interface for building byte sequences.
  """

  def __init__(self, little_endian: bool = True):
    self._buffer = BytesIO()
    self._endian = "<" if little_endian else ">"

  def u8(self, value: int) -> "Writer":
    """Write unsigned 8-bit integer (0-255)."""
    self._buffer.write(struct.pack(f"{self._endian}B", value))
    return self

  def u16(self, value: int) -> "Writer":
    """Write unsigned 16-bit integer."""
    self._buffer.write(struct.pack(f"{self._endian}H", value))
    return self

  def u32(self, value: int) -> "Writer":
    """Write unsigned 32-bit integer."""
    self._buffer.write(struct.pack(f"{self._endian}I", value))
    return self

  def u64(self, value: int) -> "Writer":
    """Write unsigned 64-bit integer."""
    self._buffer.write(struct.pack(f"{self._endian}Q", value))
    return self

  def i8(self, value: int) -> "Writer":
    """Write signed 8-bit integer (-128 to 127)."""
    self._buffer.write(struct.pack(f"{self._endian}b", value))
    return self

  def i16(self, value: int) -> "Writer":
    """Write signed 16-bit integer."""
    self._buffer.write(struct.pack(f"{self._endian}h", value))
    return self

  def i32(self, value: int) -> "Writer":
    """Write signed 32-bit integer."""
    self._buffer.write(struct.pack(f"{self._endian}i", value))
    return self

  def i64(self, value: int) -> "Writer":
    """Write signed 64-bit integer."""
    self._buffer.write(struct.pack(f"{self._endian}q", value))
    return self

  def f32(self, value: float) -> "Writer":
    """Write 32-bit float."""
    self._buffer.write(struct.pack(f"{self._endian}f", value))
    return self

  def f64(self, value: float) -> "Writer":
    """Write 64-bit double."""
    self._buffer.write(struct.pack(f"{self._endian}d", value))
    return self

  def string(self, value: str) -> "Writer":
    """Write null-terminated UTF-8 string."""
    self._buffer.write(value.encode("utf-8"))
    self._buffer.write(b"\x00")
    return self

  def raw_bytes(self, value: bytes) -> "Writer":
    """Write raw bytes."""
    self._buffer.write(value)
    return self

  def finish(self) -> bytes:
    """Return the built byte sequence."""
    return self._buffer.getvalue()


class Reader:
  """Raw byte reader.

  Reads primitive values from byte sequences.
  """

  def __init__(self, data: bytes, little_endian: bool = True):
    self._data = data
    self._offset = 0
    self._endian = "<" if little_endian else ">"

  def u8(self) -> int:
    """Read unsigned 8-bit integer."""
    if self._offset + 1 > len(self._data):
      raise ValueError(f"Not enough data for u8 at offset {self._offset}")
    value: int = struct.unpack(f"{self._endian}B", self._data[self._offset : self._offset + 1])[0]  # type: ignore[assignment]
    self._offset += 1
    return value

  def u16(self) -> int:
    """Read unsigned 16-bit integer."""
    if self._offset + 2 > len(self._data):
      raise ValueError(f"Not enough data for u16 at offset {self._offset}")
    value: int = struct.unpack(f"{self._endian}H", self._data[self._offset : self._offset + 2])[0]  # type: ignore[assignment]
    self._offset += 2
    return value

  def u32(self) -> int:
    """Read unsigned 32-bit integer."""
    if self._offset + 4 > len(self._data):
      raise ValueError(f"Not enough data for u32 at offset {self._offset}")
    value: int = struct.unpack(f"{self._endian}I", self._data[self._offset : self._offset + 4])[0]  # type: ignore[assignment]
    self._offset += 4
    return value

  def u64(self) -> int:
    """Read unsigned 64-bit integer."""
    if self._offset + 8 > len(self._data):
      raise ValueError(f"Not enough data for u64 at offset {self._offset}")
    value: int = struct.unpack(f"{self._endian}Q", self._data[self._offset : self._offset + 8])[0]  # type: ignore[assignment]
    self._offset += 8
    return value

  def i8(self) -> int:
    """Read signed 8-bit integer."""
    if self._offset + 1 > len(self._data):
      raise ValueError(f"Not enough data for i8 at offset {self._offset}")
    value: int = struct.unpack(f"{self._endian}b", self._data[self._offset : self._offset + 1])[0]  # type: ignore[assignment]
    self._offset += 1
    return value

  def i16(self) -> int:
    """Read signed 16-bit integer."""
    if self._offset + 2 > len(self._data):
      raise ValueError(f"Not enough data for i16 at offset {self._offset}")
    value: int = struct.unpack(f"{self._endian}h", self._data[self._offset : self._offset + 2])[0]  # type: ignore[assignment]
    self._offset += 2
    return value

  def i32(self) -> int:
    """Read signed 32-bit integer."""
    if self._offset + 4 > len(self._data):
      raise ValueError(f"Not enough data for i32 at offset {self._offset}")
    value: int = struct.unpack(f"{self._endian}i", self._data[self._offset : self._offset + 4])[0]  # type: ignore[assignment]
    self._offset += 4
    return value

  def i64(self) -> int:
    """Read signed 64-bit integer."""
    if self._offset + 8 > len(self._data):
      raise ValueError(f"Not enough data for i64 at offset {self._offset}")
    value: int = struct.unpack(f"{self._endian}q", self._data[self._offset : self._offset + 8])[0]  # type: ignore[assignment]
    self._offset += 8
    return value

  def f32(self) -> float:
    """Read 32-bit float."""
    if self._offset + 4 > len(self._data):
      raise ValueError(f"Not enough data for f32 at offset {self._offset}")
    value: float = struct.unpack(f"{self._endian}f", self._data[self._offset : self._offset + 4])[0]  # type: ignore[assignment]
    self._offset += 4
    return value

  def f64(self) -> float:
    """Read 64-bit double."""
    if self._offset + 8 > len(self._data):
      raise ValueError(f"Not enough data for f64 at offset {self._offset}")
    value: float = struct.unpack(f"{self._endian}d", self._data[self._offset : self._offset + 8])[0]  # type: ignore[assignment]
    self._offset += 8
    return value

  def string(self) -> str:
    """Read null-terminated UTF-8 string."""
    # Find null terminator
    null_pos = self._data.find(b"\x00", self._offset)
    if null_pos == -1:
      raise ValueError(f"No null terminator found for string at offset {self._offset}")

    # Extract string (excluding null terminator)
    string_bytes = self._data[self._offset : null_pos]
    self._offset = null_pos + 1  # Move past null terminator

    return string_bytes.decode("utf-8")

  def raw_bytes(self, n: int) -> bytes:
    """Read n raw bytes."""
    if self._offset + n > len(self._data):
      raise ValueError(f"Not enough data for {n} bytes at offset {self._offset}")
    value = self._data[self._offset : self._offset + n]
    self._offset += n
    return value

  def remaining(self) -> bytes:
    """Return all remaining unread bytes."""
    remaining = self._data[self._offset :]
    self._offset = len(self._data)
    return remaining

  def has_remaining(self) -> bool:
    """Check if there are unread bytes."""
    return self._offset < len(self._data)

  def offset(self) -> int:
    """Get current read offset."""
    return self._offset
